from __future__ import annotations

import ast
import inspect
import sys
import time
import warnings
from contextlib import ExitStack
from textwrap import dedent, indent
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

import pandas as pd

import fused_batch
from fused_batch._udf.context import ExecutionContextProtocol, context
from fused_batch._udf.state import (
    decorator_src_override_context,
    noop_decorators_context,
)
from fused_batch.models.schema import Schema
from fused_batch.models.udf import EMPTY_UDF, AnyBaseUdf
from fused_batch.models.udf._eval_result import UdfEvaluationResult
from fused_batch.models.udf.output import Output, PandasOutput
from fused_batch.warnings import FusedWarning

if TYPE_CHECKING:
    from fused_batch.models.input import BaseInput


def _add_special_kwargs(
    fn: Callable,
    args: Dict[str, Any],
    output: Output,
    context: ExecutionContextProtocol,
):
    signature = inspect.signature(fn)
    param_names = signature.parameters.keys()
    # note: we don't check that this is a kwarg on the function, so if it's the first parameter it could cause
    # a call failure.
    if "output" in param_names:
        args["output"] = output
    if "context" in param_names:
        args["context"] = context

    return args


def execute_against_sample(
    udf: AnyBaseUdf,
    input: BaseInput,
    update_schema: bool = True,
    validate_output: bool = True,
    validate_imports: Optional[bool] = None,
    **kwargs,
) -> UdfEvaluationResult:
    if udf is EMPTY_UDF:
        raise ValueError("Empty UDF cannot be evaluated. Use `set_udf` to set a UDF.")

    output = Output()
    try:
        # Define custom function in environment

        # Wrap cell in a function
        # This is believed to fix scoping issues with closures and list comprehensions
        # See also https://stackoverflow.com/a/2749806,
        # https://stackoverflow.com/q/38318370
        # {cell_text} is not itself indented because we need to make sure that every
        # line of {cell_text} is indented, not just the first.
        FN_RESULT_NAME = "__fused_fn_result"
        wrapped_cell = """\
            def __fused_exec():
            {cell_text}

                return {entrypoint}

            {FN_RESULT_NAME} = __fused_exec()
            """
        original_src = udf.code
        src = dedent(wrapped_cell).format(
            cell_text=indent(dedent(original_src), " " * 4),
            entrypoint=udf.entrypoint,
            FN_RESULT_NAME=FN_RESULT_NAME,
        )

        exec_locals = {}
        exec_globals = {"fused_batch": fused_batch}

        # Store names of modules prior to rendering Headers
        preceding_sys_modules = list(sys.modules.keys())

        # Validate import stamements correspond to valid modules
        validate_imports_whitelist(udf, validate_imports=validate_imports)

        with ExitStack() as stack:
            stack.enter_context(noop_decorators_context(True))

            # Add headers to sys.meta_path
            if udf.headers is not None:
                for header in udf.headers:
                    stack.enter_context(
                        header._register_custom_finder(preceding_sys_modules)
                    )

            exec(src, exec_globals, exec_locals)

            _fn = exec_locals[FN_RESULT_NAME]

            kwargs_to_pass = {**udf.parameters, **kwargs}
            _add_special_kwargs(
                _fn, args=kwargs_to_pass, output=output, context=context
            )
            args_to_pass = (
                input.as_udf_args() if hasattr(input, "as_udf_args") else input
            )

            # Run UDF
            time_start = time.perf_counter()
            _output = _fn(*args_to_pass, **kwargs_to_pass)
            time_end = time.perf_counter()
            time_taken_seconds = time_end - time_start

            if not validate_output:
                return _output

            if _output is not None:
                output.data = _output

    except Exception:
        raise

    new_output = _transform_output(output=output)

    if new_output is None:
        # TODO: Assumes table_schema is present, which doesn't match type
        if udf.table_schema not in [None, {}, Schema()]:
            warnings.warn(
                "UDF is configured with a schema but returns `None`. An empty schema was set for this execution.",
                FusedWarning,
            )

        return UdfEvaluationResult(
            data=None,
            sidecar=None,
            udf=udf,
            table_schema={},
            time_taken_seconds=time_taken_seconds,
        )

    # Validate the dataframe after assigning it to the output variable in user_ns
    # so the user can inspect the output if anything is wrong.
    new_output.validate_data_with_schema()

    if update_schema and udf.table_schema is None:
        udf.table_schema = new_output.table_schema

    return UdfEvaluationResult(
        data=new_output.data,
        sidecar=new_output.sidecar_output,
        udf=udf,
        table_schema=new_output.table_schema,
        time_taken_seconds=time_taken_seconds,
    )


def execute_for_decorator(udf: AnyBaseUdf) -> AnyBaseUdf:
    """Evaluate a UDF for the purpose of getting the UDF object out of it."""
    # Define custom function in environment

    # This is a stripped-down version of execute_against_sample, above.

    FN_RESULT_NAME = "__fused_fn_result"
    wrapped_cell = """\
        def __fused_exec():
        {cell_text}

            return {entrypoint}

        {FN_RESULT_NAME} = __fused_exec()
        """
    original_src = udf.code
    src = dedent(wrapped_cell).format(
        cell_text=indent(dedent(original_src), " " * 4),
        entrypoint=udf.entrypoint,
        FN_RESULT_NAME=FN_RESULT_NAME,
    )

    exec_locals = {}
    exec_globals = {"fused_batch": fused_batch}

    with ExitStack() as stack:
        stack.enter_context(decorator_src_override_context(original_src))

        # Add headers to sys.meta_path
        if udf.headers is not None:
            for header in udf.headers:
                stack.enter_context(header._register_custom_finder())

        exec(src, exec_globals, exec_locals)

        _fn = exec_locals[FN_RESULT_NAME]

        return _fn


def _transform_output(output: Output) -> PandasOutput:
    # TODO: Support PyArrow
    if isinstance(output, PandasOutput):
        return output

    if isinstance(output.data, pd.DataFrame):
        # Force all column names to be strings
        output.data.columns = [str(x) for x in output.data.columns]

        # Note that we don't pass a _new_ schema argument here
        # - If the user defined a schema onto output.table_schema, that will be
        #   preserved.
        # - If the user did not define output.table_schema, then output.table_schema
        #   will be None, and PandasOutput will default to a schema inferred from
        #   the DataFrame.
        return PandasOutput(
            data=output.data,
            table_schema=output.table_schema,
            sidecar_output=output.sidecar_output,
            skip_fused_index_validation=output.skip_fused_index_validation,
        )
    elif output.data is None:
        return None

    raise TypeError(f"Unexpected result type {type(output.data)}")


def validate_imports_whitelist(
    udf: AnyBaseUdf, validate_imports: Optional[bool] = None
):
    # Skip import validation if the option is set
    if (
        not fused_batch.options.default_validate_imports
        and validate_imports is not True
    ):
        return

    # Skip import validation if not logged in
    if not fused_batch.api.AUTHORIZATION.is_configured():
        return

    from fused_batch._global_api import get_api

    # Get the dependency whitelist from the cached API endpoint
    api = get_api()
    package_dependencies = api.dependency_whitelist()

    # Initialize a list to store the import statements
    import_statements = []

    # Parse the source code into an AST
    tree = ast.parse(udf.code)

    # Traverse the AST to find import statements
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                import_statements.append(alias.name)
        elif isinstance(node, ast.ImportFrom):
            module_name = node.module
            import_statements.append(module_name)

    # Check for unavailable modules
    header_modules = [header.module_name for header in udf.headers]
    fused_modules = ["fused"]  # assume fused is always available
    available_modules = (
        list(package_dependencies["dependency_whitelist"].keys())
        + header_modules
        + fused_modules
    )
    unavailable_modules = []
    for import_statement in import_statements:
        if import_statement.split(".", 1)[0] not in available_modules:
            unavailable_modules.append(import_statement)

    if unavailable_modules:
        raise ValueError(
            f"The following imports in the UDF might not be available: {repr(unavailable_modules)}. Please check the UDF headers and imports and try again."
        )

    # TODO: check major versions for some packages



from typing import Any, List, Tuple

from pydantic import BaseModel, ConfigDict

from .udf.common import Chunk, ChunkData, Chunks


class BaseInput(BaseModel):
    """A base class for inputs into a UDF"""

    step_config: "AnyJobStepConfig"
    """The currently executing [`step configuration`][fused.models.api.job.AnyJobStepConfig]"""

    def as_udf_args(self) -> List[Any]:
        raise NotImplementedError()


class JoinInput(BaseInput):
    left: Chunk
    """The [`chunk`][fused.models.udf.common.Chunk] containing data used as the left side of the join.
    """

    right: Chunks
    """A list of [`chunk`][fused.models.udf.common.Chunk] objects containing data used as the right side of the join.

    Only chunks whose bounding box intersects with the left chunk will be included.
    """

    def _repr_html_(self) -> str:
        return f"""
        <h2>Left</h2>
        {self.left._repr_html_()}<br/>
        <h2>Right</h2>
        {self.right._repr_html_()}
        """

    @property
    def data(self) -> Tuple[ChunkData, ChunkData]:
        """Returns the left and right data"""
        return self.left.data, self.right.concat_data()

    def as_udf_args(self) -> List[Any]:
        return [self.left.model_copy(deep=True), self.right.model_copy(deep=True)]


class JoinSingleFileInput(BaseInput):
    left: Chunk
    """The [`chunk`][fused.models.udf.common.Chunk] containing data used as the left side of the join.
    """

    right: ChunkData
    """The data used as the right side of the join.

    This will either be a [`pandas.DataFrame`][pandas.DataFrame],
    [`geopandas.GeoDataFrame`][geopandas.GeoDataFrame], or
    [`pyarrow.Table`][pyarrow.Table], depending on the type of user-defined function in
    use.
    """

    def _repr_html_(self) -> str:
        right_html = (
            f"{self.right._repr_html_()}"
            if hasattr(self.right, "_repr_html_")
            else f"{self.right}"
        )
        return f"""
        <h2>Left</h2>
        {self.left._repr_html_()}<br/>
        <h2>Right</h2>
        {right_html}
        """

    @property
    def data(self) -> Tuple[ChunkData, ChunkData]:
        """Returns the left and right data"""
        return self.left.data, self.right

    def as_udf_args(self) -> List[Any]:
        copy = self.model_copy(deep=True)
        return [copy.left, copy.right]

    model_config = ConfigDict(arbitrary_types_allowed=True)


class MapInput(BaseInput, Chunk):
    """The [`chunk`][fused.models.udf.common.Chunk] containing data used for the map operation."""

    def as_udf_args(self) -> List[Any]:
        return [self.model_copy(deep=True)]


# Workaround for a circular model reference
from .api import AnyJobStepConfig  # noqa: E402

# The job step configs return samples (the input classes below), but the input
# classes contain copies of the job step configs! This requires using a ForwardRef
# which is then resolved afterwards. This is the resolution step, so that Pydantic
# will correctly validate the type of `step_config`, above. It must be done on each
# class individually because each class has its own field information.
BaseInput.model_rebuild()
MapInput.model_rebuild()
JoinInput.model_rebuild()
JoinSingleFileInput.model_rebuild()



from datetime import timedelta
from typing import IO, Any, List, Optional

from fused_batch._global_api import get_api
from fused_batch.models.internal.job import CoerceableToJobId, RunResponse


def job_get_logs(
    job: CoerceableToJobId,
    since_ms: Optional[int] = None,
) -> List[Any]:
    """Fetch logs for a job

    Args:
        job: the identifier of a job or a `RunResponse` object.
        since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.

    Returns:
        Log messages for the given job.
    """
    api = get_api()
    return api.get_logs(job=job, since_ms=since_ms)


def job_print_logs(
    job: CoerceableToJobId,
    since_ms: Optional[int] = None,
    file: Optional[IO] = None,
) -> None:
    """Fetch and print logs for a job

    Args:
        job: the identifier of a job or a `RunResponse` object.
        since_ms: Timestamp, in milliseconds since epoch, to get logs for. Defaults to None for all logs.
        file: Where to print logs to. Defaults to sys.stdout.

    Returns:
        None
    """
    job = RunResponse.from_job_id(job)
    job.print_logs(since_ms=since_ms, file=file)


def job_tail_logs(
    job: CoerceableToJobId,
    refresh_seconds: float = 1,
    sample_logs: bool = True,
    timeout: Optional[float] = None,
    get_logs_retries: int = 1,
):
    """Continuously print logs for a job

    Args:
        job: the identifier of a job or a `RunResponse` object.
        refresh_seconds: how frequently, in seconds, to check for new logs. Defaults to 1.
        sample_logs: if true, print out only a sample of logs. Defaults to True.
        timeout: if not None, how long to continue tailing logs for. Defaults to None for indefinite.
        get_logs_retries: Number of additional retries for log requests. Defaults to 1.
    """
    api = get_api()
    return api.tail_logs(
        job=job,
        refresh_seconds=refresh_seconds,
        sample_logs=sample_logs,
        timeout=timeout,
        get_logs_retries=get_logs_retries,
    )


def job_get_status(job: CoerceableToJobId) -> RunResponse:
    """Fetch the status of a running job

    Args:
        job: the identifier of a job or a `RunResponse` object.

    Returns:
        The status of the given job.
    """
    api = get_api()
    return api.get_status(job=job)


def job_cancel(job: CoerceableToJobId) -> RunResponse:
    """Cancel an existing job

    Args:
        job: the identifier of a job or a `RunResponse` object.

    Returns:
        A new job object.
    """
    api = get_api()
    return api.cancel_job(job=job)


def job_get_exec_time(job: CoerceableToJobId) -> timedelta:
    """Determine the execution time of this job, using the logs.

    Returns:
        Time the job took. If the job is in progress, time from first to last log message is returned.
    """
    job = RunResponse.from_job_id(job)
    return job.get_exec_time()


def job_wait_for_job(
    job: CoerceableToJobId,
    poll_interval_seconds: float = 5,
    timeout: Optional[float] = None,
) -> RunResponse:
    """Block the Python kernel until this job has finished

    Args:
        poll_interval_seconds: How often (in seconds) to poll for status updates. Defaults to 5.
        timeout: The length of time in seconds to wait for the job. Defaults to None.

    Raises:
        TimeoutError: if waiting for the job timed out.

    Returns:
        The status of the given job.
    """
    api = get_api()
    return api.wait_for_job(
        job,
        poll_interval_seconds=poll_interval_seconds,
        timeout=timeout,
    )



from concurrent.futures import Future, ThreadPoolExecutor
from pathlib import Path
from typing import List
from urllib.parse import ParseResult, urlparse

import fsspec
import requests
from fsspec.implementations.dirfs import DirFileSystem
from loguru import logger

from fused_batch._global_api import get_api
from fused_batch._options import options as OPTIONS


def data_path() -> str:
    return "/tmp/fused"


def filesystem(protocol: str, **storage_options) -> fsspec.AbstractFileSystem:
    if protocol == "fd":
        # fused team directory
        api = get_api()
        if hasattr(api, "_resolve"):
            root = api._resolve("fd://")
            root_parsed = urlparse(root)
            return DirFileSystem(
                path=root, fs=fsspec.filesystem(root_parsed.scheme, **storage_options)
            )
        else:
            raise ValueError("Could not determine root of Fused team directory")
    return fsspec.filesystem(protocol, **storage_options)


def _download_requests(url: str) -> bytes:
    # this function is shared
    response = requests.get(url, headers={"User-Agent": ""})
    response.raise_for_status()
    return response.content


def _download_signed(url: str) -> bytes:
    api = get_api()
    return _download_requests(api.sign_url(url))


def _download_s3(url: str) -> bytes:
    try:
        return _download_signed(url)
    except:  # noqa E722
        s3 = filesystem("s3")
        with s3.open(url, "rb") as f:
            return f.read()


def _download_gcs(url: str) -> bytes:
    try:
        return _download_signed(url)
    except:  # noqa E722
        gcs = filesystem("gs")
        with gcs.open(url, "rb") as f:
            return f.read()


def download_folder_inner(parsed_url: ParseResult, url: str, path: str):
    if parsed_url.scheme in ["s3", "gs"]:
        api = get_api()
        all_files = api.sign_url_prefix(url)

        root_path = Path(path)
        to_remove = parsed_url.path.lstrip("/")

        def _download_single_file(filename: str, signed_url: str) -> None:
            logger.debug(f"Downloading {filename}...")
            content = _download_requests(signed_url)

            curr_path = root_path / Path(filename)
            curr_path.parent.mkdir(parents=True, exist_ok=True)

            with open(curr_path, "wb") as file:
                file.write(content)

        with ThreadPoolExecutor(max_workers=OPTIONS.max_workers) as pool:
            futures: List[Future] = []
            for filename, signed_url in all_files.items():
                if filename.startswith(to_remove):
                    filename = filename[len(to_remove) :]

                futures.append(
                    pool.submit(
                        _download_single_file,
                        filename=filename,
                        signed_url=signed_url,
                    )
                )
            pool.shutdown(wait=True)
    else:
        raise NotImplementedError(f"Unexpected URL scheme {parsed_url.scheme}")



import ast
import inspect
import re
import warnings
from datetime import timedelta
from textwrap import dedent
from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Union

from fused._udf.state import (
    decorator_src_override,
    decorator_udf_override,
    noop_decorators,
)
from fused.models import Schema
from fused.models.udf import Header
from fused.models.udf.udf import GeoPandasUdfV2
from fused.warnings import (
    FusedIgnoredWarning,
    FusedPythonVersionWarning,
    FusedTypeWarning,
    FusedUdfWarning,
)

RESERVED_UDF_PARAMETERS = {
    "self",
    "arg_list",
    "output_table",
}
"""Set of UDF parameter names that should not be used because they will cause conflicts
when instantiating the UDF to a job."""

UDF_RUN_KWARGS = {"x", "y", "z"}
"""Set of UDF keyword arguments that are also parameters to udf.run(), which would cause the
user's UDF arguments to be clobbered."""


def _extract_parameter_list(func: Callable, num_positional_args: int) -> List[str]:
    parameter_list: List[str] = []
    xyz_reserved_seen: Set = set([])

    signature = inspect.signature(func)
    for param_idx, param in enumerate(signature.parameters.values()):
        # Don't check certain
        is_reserved_positional_param = (param_idx == 0 and param.name == "dataset") or (
            param_idx == 1 and param.name == "right"
        )
        if param.name in RESERVED_UDF_PARAMETERS and not is_reserved_positional_param:
            warnings.warn(
                FusedUdfWarning(
                    f'Parameter named "{param}" is reserved and may cause conflicts. If you want to set the Fused option with the same name, provide it when instantiating the UDF but not in the parameter list.'
                )
            )

        if (
            param.kind
            in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD]
            and param.name in UDF_RUN_KWARGS
        ):
            xyz_reserved_seen.add(param.name)

        if param_idx < num_positional_args:
            # This is one of the first args for getting the input data
            continue

        if param.kind in [
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.VAR_POSITIONAL,
        ]:
            warnings.warn(
                FusedUdfWarning(
                    f"Positional-only argument (name {param.name}, kind {param.kind}) cannot be specified in UDFs"
                )
            )
        elif param.kind != inspect.Parameter.VAR_KEYWORD:
            # Must be a keyword only parameter
            # This parameter must be specified: if param.default is not inspect._empty:
            # TODO: This parameter_list is not used at all, instead the signature is called below.
            parameter_list.append(param.name)

    if len(xyz_reserved_seen) == 3:
        warnings.warn(
            FusedIgnoredWarning(
                f"Parameters named 'x', 'y' and 'z' for UDF '{func.__name__}' are "
                "reserved and will be clobbered when passing all three of them "
                "to fused.run() at the same time."
            )
        )

    return list(signature.parameters.keys())


def _parse_ttl(ttl: Optional[Union[str, int]]) -> Optional[int]:
    if ttl is None:
        return None

    if ttl == 0 or (isinstance(ttl, str) and ttl.strip() == "0"):
        return 0

    if isinstance(ttl, int):
        ttl = f"{ttl}s"

    quantifier = re.match(r"(\d+)([smhd])$", ttl.strip().lower())
    if not quantifier:
        raise ValueError(
            "Invalid TTL format: Use a number followed by one of 's' (seconds), 'm' (minutes), 'h' (hours), or 'd' (days)."
        )

    value, unit = int(quantifier.group(1)), quantifier.group(2)
    delta_map = {
        "s": timedelta(seconds=value),
        "m": timedelta(minutes=value),
        "h": timedelta(hours=value),
        "d": timedelta(days=value),
    }
    return int(delta_map[unit].total_seconds())


# Note: If the signature of this function changes, the signature of the
# specialized versions of it below must also change. This pattern is used
# so that the internal function can be reused (pass in different _udf_cls),
# but the type hinting in e.g. VS Code is correct. Importantly, this type
# hinting includes the keyword parameters. Otherwise the way to express
# the type to VS Code is either not possible or too complicated.
def _udf_internal(
    fn: Optional[Callable] = None,
    *,
    _udf_cls: type[GeoPandasUdfV2],
    _num_positional_args: int,
    schema: Union[Schema, Dict, str, None] = None,
    name: Union[str, None] = None,
    cache_max_age: Optional[str] = None,
    default_parameters: Optional[Dict[str, Any]] = None,
    headers: Optional[Sequence[Union[str, Header]]] = None,
):
    def _internal_decorator_wrapper(func):
        if noop_decorators.get():
            return func

        entrypoint = func.__name__
        fn_name = name or entrypoint

        _src = decorator_src_override.get() or inspect.getsource(func)
        _src = dedent(_src)
        # src = _strip_decorators(_src)
        _src, original_headers = _strip_decorator_params(_src, cache_max_age)

        src = _src

        parameter_list = _extract_parameter_list(
            func, num_positional_args=_num_positional_args
        )

        table_schema = schema  # may be null
        if table_schema is not None:
            if isinstance(table_schema, str):
                table_schema = Schema.from_string(table_schema)
            elif isinstance(table_schema, Schema):
                warnings.warn(
                    FusedTypeWarning(
                        "Passing table_schema as type Schema may not work on the backend. Instead pass a JSON string literal."
                    )
                )
            elif not isinstance(table_schema, Schema):
                # Schema really should not be of type Schema, since that cannot be
                # run on the backend.
                table_schema = Schema.model_validate(table_schema)

        resolved_headers = headers
        override_udf = decorator_udf_override.get()
        if override_udf and override_udf.headers:
            # model_dump is needed because the headers are really the wrong type
            resolved_headers = [h.model_dump() for h in override_udf.headers]

        new_udf = _udf_cls(
            code=src.strip("\n"),
            name=fn_name,
            entrypoint=entrypoint,
            cache_max_age=_parse_ttl(cache_max_age),
            table_schema=table_schema,
            parameters=default_parameters or {},
            headers=resolved_headers,
            original_headers=original_headers,
        )
        new_udf._parameter_list = parameter_list
        new_udf._nested_callable = func
        return new_udf

    if fn is not None:
        return _internal_decorator_wrapper(fn)
    else:
        return _internal_decorator_wrapper


# Specializations of _udf_internal. Note we return Callable so that type hinting
# is right when using "@fused.udf()" -- but this does mean it's wrong when
# using "@fused.udf".
def udf(
    fn: Optional[Callable] = None,
    *,
    schema: Union[Schema, Dict, None] = None,
    name: Optional[str] = None,
    cache_max_age: Optional[str] = None,
    default_parameters: Optional[Dict[str, Any]] = None,
    headers: Optional[Sequence[Union[str, Header]]] = None,
) -> Callable[..., GeoPandasUdfV2]:
    """A decorator that transforms a function into a Fused UDF.

    Args:
        schema: The schema for the DataFrame returned by the UDF. The schema may be either
            a string (in the form `"field_name:DataType field_name2:DataType"`, or as JSON),
            as a Python dictionary representing the schema, or a `Schema` model object.

            Defaults to None, in which case a schema must be evaluated by calling `run_local`
            for a job to be able to write output. The return value of `run_local` will also
            indicate how to include the schema in the decorator so `run_local` does not need
            to be run again.
        name: The name of the UDF object. Defaults to the name of the function.
        cache_max_age: The maximum age when returning a result from the cache.
        default_parameters: Parameters to embed in the UDF object, separately from the arguments
            list of the function. Defaults to None for empty parameters.
        headers: A list of files to include as modules when running the UDF. For example,
            when specifying `headers=['my_header.py']`, inside the UDF function it may be
            referenced as:

            ```py
            import my_header
            my_header.my_function()
            ```

            Defaults to None for no headers.
    Returns:
        A callable that represents the transformed UDF. This callable can be used
        within GeoPandas workflows to apply the defined operation on geospatial data.

    Examples:
        To create a simple UDF that calls a utility function to calculate the area of geometries in a GeoDataFrame:

        ```py
        @fused.udf
        def udf(bbox, table_path="s3://fused-asset/infra/building_msft_us"):
            ...
            gdf = table_to_tile(bbox, table=table_path)
            return gdf
        ```
    """
    return _udf_internal(
        fn=fn,
        _udf_cls=GeoPandasUdfV2,
        _num_positional_args=2,
        schema=schema,
        name=name,
        cache_max_age=cache_max_age,
        default_parameters=default_parameters,
        headers=headers,
    )


def _strip_decorator_params(src: str, cache_max_age=None) -> str:
    """Remove all parameter from decorator declaration."""
    # TODO: use ast by line number to not use ast.unparse

    # TODO: This requires Python 3.9 because ast.unparse was exposed in that release.
    # https://docs.python.org/3/whatsnew/3.9.html#ast We may be able to backport it?
    if not hasattr(ast, "unparse"):
        warnings.warn(
            FusedPythonVersionWarning(
                "Decorators will not be removed from UDF source because ast.unparse is not available in this version of Python"
            )
        )
        return src, ""

    tree = ast.parse(src)

    original_headers = ""
    changes_made = False

    def remove_params_from_decorator(node):
        nonlocal changes_made
        if isinstance(node, ast.Call):
            if (
                isinstance(node.func, ast.Attribute)
                and node.func.value.id == "fused"
                and node.func.attr == "udf"
            ):
                keywords = []
                for each_keyword in node.keywords:
                    if each_keyword.arg == "headers":
                        nonlocal original_headers
                        original_headers = ast.unparse(each_keyword.value)
                        changes_made = True
                    elif each_keyword.arg == "cache_max_age":
                        # if cache_max_age was passed as a variable, replace with its literal
                        if not isinstance(each_keyword.value, ast.Constant):
                            each_keyword.value = ast.Constant(value=cache_max_age)
                        changes_made = True
                        keywords.append(each_keyword)
                node.keywords = keywords
        return node

    # Remove parameters from the fused decorator definition
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            new_decorators = []
            for decorator in node.decorator_list:
                if isinstance(decorator, ast.Call):
                    new_decorators.append(remove_params_from_decorator(decorator))
                else:
                    new_decorators.append(decorator)

    if changes_made:
        modified_src = ast.unparse(tree)

        return modified_src, original_headers
    else:
        # Avoid unparsing if not absolutely necessary. This is because of e.g.
        # PEP 701 causing differences between Python 3.11/3.12 rendering of f-strings.
        return src, ""



import datetime
from typing import Optional

from pydantic import BaseModel


class ListDetails(BaseModel):
    url: str
    file_name: str
    is_directory: bool = False
    size: Optional[int] = None
    """Size in bytes"""
    last_modified: Optional[datetime.datetime] = None



# ruff: noqa: F401

from fused_batch._auth import AUTHORIZATION

from ._public_api import (
    job_cancel,
    job_get_exec_time,
    job_get_logs,
    job_get_status,
    job_print_logs,
    job_tail_logs,
    job_wait_for_job,
)
from .api import FusedAPI
from .credentials import NotebookCredentials, access_token, auth_scheme, logout
from .docker_api import FusedDockerAPI
from .docker_http_api import FusedDockerHTTPAPI



# ruff: noqa: F401

from .job import (
    CustomJobConfig,
    create_directory_and_zip,
    generate_meta_json,
    generate_readme,
)
from .udf import (
    extract_parameters,
    stringify_headers,
    stringify_input,
    stringify_named_params,
    stringify_output,
    structure_params,
)



from typing import Optional

from pydantic import BaseModel, PrivateAttr, StrictStr


class FusedProjectAware(BaseModel):
    _project_url: Optional[StrictStr] = PrivateAttr()
    """Project base path that table names will be resolved relative to"""



import ast
from pathlib import Path
from typing import Optional

from fused_batch._str_utils import table_to_name

from ..internal import (
    DatasetInput,
    DatasetInputV2,
    DatasetInputV2Type,
    DatasetOutput,
    DatasetOutputV2,
)


def stringify_named_params(params):
    return [k + "=" + repr(v) for k, v in params.items()]


def structure_params(params, separator=", "):
    return separator.join([param for param in params])


def stringify_headers(headers):
    if not headers:
        return "[]"
    else:
        for header in headers:
            if hasattr(header, "source_file"):
                delattr(header, "source_file")
        return str(headers)


def stringify_input(input) -> str:
    if isinstance(input, str):
        return repr(input)
    elif isinstance(input, DatasetInputV2):
        op_fn: Optional[str] = None
        if input.operation == DatasetInputV2Type.ZIP:
            op_fn = "fused_batch.zip_tables"
        elif input.operation == DatasetInputV2Type.UNION:
            op_fn = "fused_batch.union_tables"

        any_has_sidecar = any([t.read_sidecar_files for t in input.tables])
        # TODO: cache locally

        if len(input.tables) == 1 and not any_has_sidecar:
            return repr(str(input.tables[0].url))
        elif op_fn:
            if any_has_sidecar:
                read_sidecar = repr(
                    [table_to_name(t) for t in input.tables if t.read_sidecar_files]
                )
                read_sidecar_fragment = f", read_sidecar={read_sidecar}"
            else:
                read_sidecar_fragment = ""
            return f"{op_fn}({repr([str(t.url) for t in input.tables])}{read_sidecar_fragment})"
    elif isinstance(input, DatasetInput):
        # Note this is rewriting from V1 to V2 style

        base_path = input.base_path.rstrip("/")

        if len(input.tables) == 1:
            return repr(f"{base_path}/{input.tables[0]}")

        return f"fused_batch.zip_tables([{', '.join([repr(f'{base_path}/{t}') for t in input.tables])}])"

    return repr(input)


def stringify_output(output) -> Optional[str]:
    if isinstance(output, DatasetOutputV2):
        return repr(output.url) if output.url is not None else None

    if isinstance(output, DatasetOutput):
        # It could be tricky to render a DatasetOutput, so special case the empty output
        if output.base_path is None and output.table is None:
            return None
        elif output.base_path is not None and output.table is not None:
            return repr(Path(output.base_path) / output.table)
        else:
            return repr(output.base_path or output.table)

    return repr(output)


def extract_parameters(src):
    # Parse the input string into an AST (Abstract Syntax Tree)
    parsed_ast = ast.parse(src)

    all_parameters = []
    named_parameters = {}

    # Find all function definitions in the AST
    function_defs = [
        node
        for node in ast.walk(parsed_ast)
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))
    ]
    # Assume the first function is the target function
    target_function = function_defs[0]

    for arg in target_function.args.args:
        param_name = arg.arg
        # named_parameters[param_name] = None  # Initialize with None
        all_parameters.append(param_name)

    for keyword in target_function.args.kwonlyargs:
        param_name = keyword.arg
        # named_parameters[param_name] = None  # Initialize with None

    for param, default in zip(
        target_function.args.args[
            len(target_function.args.args) - len(target_function.args.defaults) :
        ],
        target_function.args.defaults,
    ):
        # Extract the default value if it's a string, number, or None
        # TODO: Handle more types
        if isinstance(
            default,
            (
                ast.Str,
                ast.Num,
                ast.NameConstant,
                ast.Tuple,
                ast.List,
                ast.Dict,
                ast.Set,
            ),
        ):
            named_parameters[param.arg] = ast.literal_eval(default)

    positional_parameters = [
        param for param in all_parameters if param not in named_parameters
    ]

    return positional_parameters, named_parameters



import warnings
from typing import Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import pyarrow as pa
from pydantic import (
    BaseModel,
    ConfigDict,
    ValidationInfo,
    field_validator,
    model_validator,
)

from fused_batch.warnings import FusedUdfWarning

from ..schema import INDEX_COLUMN_NAME, Schema


class Output(BaseModel):
    skip_fused_index_validation: bool = False
    data: Union[pa.Table, pd.DataFrame, gpd.GeoDataFrame, None] = None
    table_schema: Optional[Schema] = None
    # TODO: sidecar_output can be a path too, right?
    sidecar_output: Optional[bytes] = None

    @field_validator("data", mode="before")
    @classmethod
    def coerce_none_to_empty_dataframe(cls, v):
        if v is None:
            v = pd.DataFrame({"fused_index": pd.Series([], dtype=np.uint32)})

        return v

    @field_validator("data")
    @classmethod
    def validate_data(cls, v, info: ValidationInfo):
        if info.data.get("skip_fused_index_validation", False):
            return v
        fused_index_msg = "`fused_index` must be a column in `output.data`."
        if isinstance(v, pd.DataFrame):
            assert "fused_index" in v.columns, fused_index_msg

        elif isinstance(v, pa.Table):
            assert "fused_index" in v.column_names, fused_index_msg

        return v

    def validate_data_with_schema(self, *, strict: bool = False):
        if strict:
            raise NotImplementedError

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)


class PandasOutput(Output):
    data: Union[gpd.GeoDataFrame, pd.DataFrame]
    """The output DataFrame or GeoDataFrame."""

    table_schema: Schema
    """A schema describing the output table."""

    @model_validator(mode="before")
    @classmethod
    def initialize_table_schema(cls, values):
        """If table_schema is not passed, initialize one from the data argument"""
        if values.get("table_schema") is None:
            assert "data" in values, "`data` was not passed into PandasOutput"
            values["table_schema"] = Schema.from_dataframe(values["data"])

        return values

    def validate_data_with_schema(self, *, strict: bool = False):
        super().validate_data_with_schema(strict=strict)
        if not self.skip_fused_index_validation:
            assert (
                INDEX_COLUMN_NAME in self.data.columns
            ), f"`{INDEX_COLUMN_NAME}` should be a column on `output.data`"

            # if fused index is wrong type, it's an error (or warning?)
            index_dtype = self.data[INDEX_COLUMN_NAME].dtype
            if index_dtype.kind == "i" or index_dtype.kind == "u":
                if (self.data[INDEX_COLUMN_NAME] < 0).any():
                    # This only makes sense for signed
                    warnings.warn(
                        f"`{INDEX_COLUMN_NAME}` has negative values, which may cause errors.",
                        FusedUdfWarning,
                    )
            else:
                warnings.warn(
                    f"Expected `{INDEX_COLUMN_NAME}` to be an integer, but it was {index_dtype}",
                    FusedUdfWarning,
                )

        # if there's any difference between field names in the schema and the df names,
        # error
        field_names = {field.name for field in self.table_schema.fields}
        field_names.add(INDEX_COLUMN_NAME)

        df_column_diff = set([str(x) for x in self.data.columns]).difference(
            field_names
        )
        if df_column_diff:
            raise ValueError(
                f"Additional columns specified that were not in the schema: {', '.join(df_column_diff)}"
            )

        if isinstance(self.data, gpd.GeoDataFrame):
            # Remove geometry before validating
            non_geo_columns = [
                name
                for name in self.data.columns
                if name != self.data._geometry_column_name
            ]
            non_geo_df = self.data[non_geo_columns]

            non_geo_schema = Schema(
                fields=[
                    field
                    for field in self.table_schema.fields
                    if field.name != self.data._geometry_column_name
                ]
            )

            pa.Table.from_pandas(non_geo_df, schema=non_geo_schema.to_arrow())

        else:
            # Lax parsing; this allows anything where pyarrow _can_ cast between the two
            # types. For example this allows Uint64 -> Float64, but not Uint64 ->
            # String.
            # If validated, this should not error
            pa.Table.from_pandas(self.data, schema=self.table_schema.to_arrow())


class PyArrowOutput(Output):
    data: pa.Table
    """The output Table."""

    def validate_data_with_schema(self, *, strict: bool = False):
        super().validate_data_with_schema(strict=strict)
        assert (
            INDEX_COLUMN_NAME in self.data.column_names
        ), f"`{INDEX_COLUMN_NAME}` should be a column on `output.data`"

        index_type = self.data.field(INDEX_COLUMN_NAME)
        if (
            pa.int64() == index_type
            or pa.int32() == index_type
            or pa.int16() == index_type
            or pa.int8() == index_type
            or pa.uint64() == index_type
            or pa.uint32() == index_type
            or pa.uint16() == index_type
            or pa.uint8() == index_type
        ):
            # TODO: Check for negative indexes here
            pass
        else:
            warnings.warn(
                f"Expected `{INDEX_COLUMN_NAME}` to be an integer, but it was {index_type}",
                FusedUdfWarning,
            )



"""Models to describe objects for input/output of a UDF
"""

# ruff: noqa: F401

from .base_udf import BaseUdf
from .common import Chunk, ChunkData, ChunkMetadata, Chunks
from .header import Header
from .udf import EMPTY_UDF, AnyBaseUdf, GeoPandasUdfV2, RootAnyBaseUdf



import os
import time
from pathlib import Path
from typing import Callable

from ._impl._download_impl import (
    _download_gcs,
    _download_requests,
    _download_s3,
    data_path,
)


def create_path(path: str, mkdir: bool = True) -> str:
    # Deprecated name...
    return file_path(file_path=path, mkdir=mkdir)


def file_path(file_path: str, mkdir: bool = True) -> str:
    """Creates a directory in a predefined temporary directory.

    This gives users the ability to manage directories during the execution of a UDF. It takes a relative file_path,
    creates the corresponding directory structure, and returns its absolute path.

    This is useful for UDFs that temporarily store intermediate results as files,
    such as when writing intermediary files to disk when processing large datasets.
    file_path ensures that necessary directories exist.

    Args:
        file_path: The file path to locate.
        mkdir: If True, create the directory if it doesn't already exist. Defaults to True.

    Returns:
        The located file path.
    """
    # TODO: Move this onto the context object or use the context object
    # TODO: Return Path objects rather than converting to str
    global_path = data_path() + "/"

    tmp = file_path.split("/")
    if tmp[-1].rfind(".") < 0:
        folder = Path(global_path + "/" + file_path)
    else:
        folder = Path(global_path + "/" + "/".join(tmp[:-1]))
    if mkdir:
        folder.mkdir(parents=True, exist_ok=True)
    return str(Path(global_path + "/" + file_path))


def download(url: str, file_path: str) -> str:
    """Download a file.

    May be called from multiple processes with the same inputs to get the same result.

    Fused runs UDFs from top to bottom each time code changes. This means objects in the UDF are recreated each time, which can slow down a UDF that downloads files from a remote server.

    💡 Downloaded files are written to a mounted volume shared across all UDFs in an organization. This means that a file downloaded by one UDF can be read by other UDFs.

    Fused addresses the latency of downloading files with the download utility function. It stores files in the mounted filesystem so they only download the first time.

    💡 Because a Tile UDF runs multiple chunks in parallel, the download function sets a signal lock during the first download attempt, to ensure the download happens only once.

    Args:
        url: The URL to download.
        file_path: The local path where to save the file.

    Returns:
        The function downloads the file only on the first execution, and returns the file path.

    Examples:
        ```python
        @fused.udf
        def geodataframe_from_geojson():
            import geopandas as gpd
            url = "s3://sample_bucket/my_geojson.zip"
            path = fused.core.download(url, "tmp/my_geojson.zip")
            gdf = gpd.read_file(path)
            return gdf
        ```

    """
    from urllib.parse import urlparse

    from loguru import logger

    file_path = file_path.strip("/")

    # Cache in mounted drive if available & writable, else cache in /tmp
    base_path = data_path()

    # Download directory
    file_full_path = Path(base_path) / file_path
    file_full_path.parent.mkdir(parents=True, exist_ok=True)

    def _download():
        parsed_url = urlparse(url)
        logger.debug(f"Downloading {url} -> {file_full_path}")

        if parsed_url.scheme == "s3":
            content = _download_s3(url)
        elif parsed_url.scheme == "gs":
            content = _download_gcs(url)
        else:
            if parsed_url.scheme not in ["http", "https"]:
                logger.debug(f"Unexpected URL scheme {parsed_url.scheme}")
            content = _download_requests(url)

        with open(file_full_path, "wb") as file:
            file.write(content)

    _run_once(signal_name=file_path, fn=_download)

    return file_full_path


def _run_once(signal_name: str, fn: Callable) -> None:
    """Run a function once, waiting for another process to run it if in progress.

    Args:
        signal_key: A relative key for signalling done status. Files are written using `file_path` and this key to deduplicate runs.
        fn: A function that will be run once.
    """
    from loguru import logger

    path_in_progress = Path(file_path(signal_name + ".in_progress"))
    path_done = Path(file_path(signal_name + ".done"))
    path_error = Path(file_path(signal_name + ".error"))

    def _wait_for_file_done():
        logger.debug(f"Waiting for {signal_name}")
        while not path_done.exists() and not path_error.exists():
            time.sleep(1)
        if path_error.exists():
            os.remove(str(path_in_progress))
            os.remove(str(path_error))
            raise ValueError(f"{signal_name} failed in another chunk. Try again.")
        logger.info(f"already cached ({signal_name}).")

    if path_in_progress.is_file():
        _wait_for_file_done()
    else:
        try:
            with open(path_in_progress, "x") as file:
                file.write("requesting")
        except FileExistsError:
            _wait_for_file_done()
            return
        logger.debug(f"Running fn -> {signal_name}")

        try:
            fn()
        except:
            with open(path_error, "w") as file:
                file.write("done")
            raise

        with open(path_done, "w") as file:
            file.write("done")
        logger.info(f"waited successfully ({signal_name}).")



import warnings
from functools import partial
from typing import Any, Callable, Coroutine, Dict, Literal, Optional, Union, overload

import geopandas as gpd
import pandas as pd
import shapely
import xarray as xr

from .core import (
    run_file,
    run_file_async,
    run_shared_file,
    run_shared_file_async,
    run_shared_tile,
    run_shared_tile_async,
    run_tile,
    run_tile_async,
)
from .core._impl._realtime_ops_impl import default_run_engine
from .core._impl._reimports import GeoPandasUdfV2, UdfAccessToken, UdfJobStepConfig

ResultType = Union[xr.Dataset, pd.DataFrame, gpd.GeoDataFrame]


@overload
def run(
    email_or_udf_or_token: Union[
        str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken
    ],
    /,
    udf_name: Optional[str],
    *,
    udf: Optional[GeoPandasUdfV2],
    job_step: Optional[UdfJobStepConfig],
    token: Optional[str],
    udf_email: Optional[str],
    x: Optional[int],
    y: Optional[int],
    z: Optional[int],
    _lat: Optional[float],
    _lng: Optional[float],
    bbox: Union[gpd.GeoDataFrame, shapely.Geometry, None],
    sync: bool = False,
    engine: Optional[Literal["realtime", "batch", "local"]],
    type: Optional[Literal["tile", "file"]],
    parameters: Optional[Dict[str, Any]] = None,
    **kw_parameters,
) -> Coroutine[ResultType, None, None]:
    ...


@overload
def run(
    email_or_udf_or_token: Union[str, None, UdfJobStepConfig, GeoPandasUdfV2],
    /,
    udf_name: Optional[str],
    *,
    udf: Optional[GeoPandasUdfV2],
    job_step: Optional[UdfJobStepConfig],
    token: Optional[str],
    udf_email: Optional[str],
    x: Optional[int],
    y: Optional[int],
    z: Optional[int],
    _lat: Optional[float],
    _lng: Optional[float],
    bbox: Union[gpd.GeoDataFrame, shapely.Geometry, None],
    sync: bool,
    engine: Optional[Literal["realtime", "batch", "local"]],
    type: Optional[Literal["tile", "file"]],
    parameters: Optional[Dict[str, Any]] = None,
    **kw_parameters,
) -> ResultType:
    ...


def run(
    email_or_udf_or_token: Union[
        str, None, UdfJobStepConfig, GeoPandasUdfV2, UdfAccessToken
    ] = None,
    /,
    udf_name: Optional[str] = None,
    *,
    udf: Optional[GeoPandasUdfV2] = None,
    job_step: Optional[UdfJobStepConfig] = None,
    token: Optional[str] = None,
    udf_email: Optional[str] = None,
    x: Optional[int] = None,
    y: Optional[int] = None,
    z: Optional[int] = None,
    _lat: Optional[float] = None,
    _lng: Optional[float] = None,
    bbox: Union[gpd.GeoDataFrame, shapely.Geometry, None] = None,
    sync: bool = True,
    engine: Optional[Literal["realtime", "batch", "local"]] = None,
    type: Optional[Literal["tile", "file"]] = None,
    parameters: Optional[Dict[str, Any]] = None,
    **kw_parameters,
):
    """
    Executes a user-defined function (UDF) with various execution and input options.

    This function supports executing UDFs in different environments (realtime, batch, local),
    with different types of inputs (tile coordinates, geographical bounding boxes, etc.), and
    allows for both synchronous and asynchronous execution. It dynamically determines the execution
    path based on the provided parameters.

    Args:
        email_or_udf_or_token: A string that can either be an email, a UDF token, or a direct
            reference to a UDF object. It can also be a UdfJobStepConfig object for detailed
            configuration, or None to specify UDF details in other parameters.
        udf_name: The name of the UDF to execute.
        udf: A GeoPandasUdfV2 object for direct execution.
        job_step: A UdfJobStepConfig object for detailed execution configuration.
        token: A token representing a shared UDF.
        udf_email: The email associated with the UDF.
        x, y, z: Tile coordinates for tile-based UDF execution.
        _lat, _lng: Latitude and longitude for tile-based UDF execution.
        bbox: A geographical bounding box (as a GeoDataFrame or shapely Geometry) defining the area of interest.
        sync: If True, execute the UDF synchronously. If False, execute asynchronously.
        engine: The execution engine to use ('realtime', 'batch', or 'local').
        type: The type of UDF execution ('tile' or 'file').
        parameters: Additional parameters to pass to the UDF.
        **kw_parameters: Additional parameters to pass to the UDF.

    Raises:
        ValueError: If the UDF is not specified or is specified in more than one way.
        TypeError: If the first parameter is not of an expected type.
        Warning: Various warnings are issued for ignored parameters based on the execution path chosen.

    Returns:
        The result of the UDF execution, which varies based on the UDF and execution path.

    Examples:


        Run a UDF saved in the Fused system:
        ```py
        fused.run(udf_email="username@fused.io", udf_name="my_udf_name")
        ```

        Run a UDF saved in GitHub:
        ```py
        loaded_udf = fused.load("https://github.com/fusedio/udfs/tree/main/public/Building_Tile_Example")
        fused.run(udf=loaded_udf, bbox=bbox)
        ```

        Run a UDF saved in a local directory:
        ```py
        loaded_udf = fused.load("/Users/local/dir/Building_Tile_Example")
        fused.run(udf=loaded_udf, bbox=bbox)
        ```

    Note:
        This function dynamically determines the execution path and parameters based on the inputs.
        It is designed to be flexible and support various UDF execution scenarios.
    """
    if email_or_udf_or_token is not None:
        if udf is not None:
            warnings.warn(
                "udf parameter is being ignored in favor of the first positional parameter.",
            )
        if job_step is not None:
            warnings.warn(
                "job_step parameter is being ignored in favor of the first positional parameter.",
            )
        if token is not None:
            warnings.warn(
                "token parameter is being ignored in favor of the first positional parameter.",
            )
        if udf_email is not None:
            warnings.warn(
                "udf_email parameter is being ignored in favor of the first positional parameter.",
            )

        if isinstance(email_or_udf_or_token, UdfJobStepConfig):
            job_step = email_or_udf_or_token
        elif isinstance(email_or_udf_or_token, GeoPandasUdfV2):
            udf = email_or_udf_or_token
        elif isinstance(email_or_udf_or_token, UdfAccessToken):
            token = email_or_udf_or_token.token
        elif isinstance(email_or_udf_or_token, str):
            if "/" in email_or_udf_or_token:
                udf_email, udf_name = email_or_udf_or_token.split("/", maxsplit=1)
            elif "@" in email_or_udf_or_token:
                udf_email = email_or_udf_or_token
            else:
                # TODO: no way to specify only UDF name
                token = email_or_udf_or_token
        else:
            raise TypeError(
                "Could not detect UDF from first parameter. It should be a string, UdfJobStepConfig, or BaseUdf object."
            )

    ways_of_specifying_udfs = sum([bool(x) for x in [udf, job_step, token, udf_email]])
    if ways_of_specifying_udfs == 0:
        raise ValueError("No UDF specified")
    if ways_of_specifying_udfs != 1:
        raise ValueError("UDF was specified in more than one way, which is invalid")

    local_tile_bbox: Optional[gpd.GeoDataFrame] = None
    if bbox is not None:
        if x is not None or y is not None or z is not None:
            warnings.warn("x, y, z arguments will be ignored in favor of bbox")
        if _lat is not None or _lng is not None:
            warnings.warn("_lat, _lng arguments will be ignored in favor of bbox")

        if (
            "x" in bbox.columns
            and "y" in bbox.columns
            and "z" in bbox.columns
            and len(bbox) == 1
        ):
            x, y, z = bbox.iloc[0][["x", "y", "z"]]
            local_tile_bbox = bbox.copy(deep=True)
        else:
            # bbox is not an x,y,z tile bounds from the Fused system, so try to find the appropriate bounds
            import mercantile

            if isinstance(bbox, gpd.GeoDataFrame):
                tile_bounds = bbox.total_bounds
            else:
                tile_bounds = shapely.bounds(bbox)

            tile = mercantile.bounding_tile(*tile_bounds)
            x, y, z = tile.x, tile.y, tile.z
            local_tile_bbox = gpd.GeoDataFrame(
                {"x": [x], "y": [y], "z": [z]}, geometry=[shapely.box(*tile_bounds)]
            )
    elif _lat is not None and _lng is not None:
        if z is None:
            raise ValueError(
                "Cannot compute x, y tile coordinates from _lat, _lng without z. You must specify z."
            )
        if x is not None or y is not None:
            warnings.warn("x and y arguments will be ignored in favor of _lat, _lng")

        import mercantile

        tile = mercantile.tile(lng=_lng, lat=_lat, zoom=z)
        tile_bounds = mercantile.bounds(tile)
        x, y, z = tile.x, tile.y, tile.z
        local_tile_bbox = gpd.GeoDataFrame(
            {"x": [x], "y": [y], "z": [z]}, geometry=[shapely.box(*tile_bounds)]
        )
    elif _lat is not None or _lng is not None:
        warnings.warn("lat, lng arguments will be ignored because one of them is None")
    elif x is not None and y is not None and z is not None:
        import mercantile

        tile_bounds = mercantile.bounds(x, y, z)
        local_tile_bbox = gpd.GeoDataFrame(
            {"x": [x], "y": [y], "z": [z]}, geometry=[shapely.box(*tile_bounds)]
        )
    elif x is not None or y is not None or z is not None:
        warnings.warn("x, y, z arguments will be ignored because one of them is None")

    if x is not None and y is not None and z is not None:
        if type is None:
            type = "tile"
        elif type != "tile":
            warnings.warn(
                "x, y, z specified but UDF type is not 'tile', so they will be ignored",
            )
    else:
        if type is None:
            type = "file"
        elif type != "file":
            raise ValueError(
                "x, y, z not specified but type is 'tile', which is an invalid configuration. You must specify x, y, and z."
            )

    if udf is not None:
        job_step = UdfJobStepConfig(udf=udf)

    parameters = {
        **kw_parameters,
        **(parameters if parameters is not None else {}),
    }

    dispatch: dict[
        tuple[
            Literal["sync", "async"],
            Literal["tile", "file"],
            Literal["saved", "token", "local_job_step"],
            Literal["realtime", "local", "batch"],
        ],
        Optional[Callable],
    ] = {
        # Saved UDF
        ("sync", "tile", "saved", "realtime"): partial(
            run_tile, udf_email, udf_name, x=x, y=y, z=z, **parameters
        ),
        ("async", "tile", "saved", "realtime"): partial(
            run_tile_async, udf_email, udf_name, x=x, y=y, z=z, **parameters
        ),
        ("sync", "file", "saved", "realtime"): partial(
            run_file, udf_email, udf_name, **parameters
        ),
        ("async", "file", "saved", "realtime"): partial(
            run_file_async, udf_email, udf_name, **parameters
        ),
        # shared UDF token
        ("sync", "tile", "token", "realtime"): partial(
            run_shared_tile, token, x=x, y=y, z=z, **parameters
        ),
        ("async", "tile", "token", "realtime"): partial(
            run_shared_tile_async, token, x=x, y=y, z=z, **parameters
        ),
        ("sync", "file", "token", "realtime"): partial(
            run_shared_file, token, **parameters
        ),
        ("async", "file", "token", "realtime"): partial(
            run_shared_file_async, token, **parameters
        ),
        # Local job step, which includes locally held code
        ("sync", "tile", "local_job_step", "realtime"): lambda: job_step.run_tile(
            x=x, y=y, z=z, **parameters
        ),
        ("async", "tile", "local_job_step", "realtime"): None,
        ("sync", "file", "local_job_step", "realtime"): lambda: job_step.run_file(
            **parameters
        ),
        ("async", "file", "local_job_step", "realtime"): None,
        ("sync", "tile", "local_job_step", "local"): lambda: job_step.run_local(
            local_tile_bbox, **parameters
        ),
        ("async", "tile", "local_job_step", "local"): lambda: job_step.run_local(
            local_tile_bbox, **parameters
        ),
        ("sync", "file", "local_job_step", "local"): lambda: job_step.run_local(
            **parameters
        ),
        ("async", "file", "local_job_step", "local"): lambda: job_step.run_local(
            **parameters
        ),
        ("sync", "tile", "local_job_step", "batch"): None,
        ("sync", "file", "local_job_step", "batch"): lambda: job_step.set_udf(
            job_step.udf, parameters=parameters
        ).run_remote(),
    }

    udf_storage = (
        "saved"
        if udf_email is not None
        else (
            "token"
            if token is not None
            else ("local_job_step" if job_step is not None else None)
        )
    )
    if udf_storage is None:
        raise ValueError("No UDF specified")

    if engine is None:
        if udf_storage in ["token", "saved"]:
            engine = "realtime"
        else:
            engine = default_run_engine()

    dispatch_params = ("sync" if sync else "async", type, udf_storage, engine)

    # Ellipsis is the sentinal value for not found in the dictionary at all
    fn = dispatch.get(dispatch_params, ...)

    if fn is Ellipsis:
        if udf_storage == "token" and engine != "realtime":
            raise ValueError("UDF tokens can only be called on the realtime engine.")
        elif udf_storage == "saved" and engine != "realtime":
            raise ValueError(
                "Saved UDFs can only be called on the realtime engine. To use another engine, load the UDF locally first."
            )
        else:
            raise ValueError(
                f"Could not determine how to call with settings: {dispatch_params}"
            )
    if fn is None:
        raise ValueError(f"Call type is not yet implemented: {dispatch_params}")

    return fn()



from typing import Any, List, Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import pyarrow as pa
import xarray as xr
from pydantic import BaseModel, ConfigDict

from fused_batch._formatter.formatter_eval_result import fused_eval_result_repr
from fused_batch.models.schema import Schema
from fused_batch.models.udf import AnyBaseUdf


class UdfEvaluationResult(BaseModel):
    data: Union[
        gpd.GeoDataFrame,
        pd.DataFrame,
        pa.Table,
        xr.Dataset,
        xr.DataArray,
        np.ndarray,
        None,
    ] = None
    sidecar: Optional[bytes] = None

    udf: Optional[AnyBaseUdf] = None
    table_schema: Optional[Schema] = None

    time_taken_seconds: float

    stdout: Optional[str] = None
    stderr: Optional[str] = None
    error_message: Optional[str] = None
    error_lineno: Optional[int] = None

    def _repr_html_(self) -> str:
        return fused_eval_result_repr(self)

    model_config = ConfigDict(arbitrary_types_allowed=True)


class MultiUdfEvaluationResult(BaseModel):
    udf_results: List[Union[UdfEvaluationResult, Any]]

    def _repr_html_(self) -> str:
        # Aggregate reprs
        result_reprs = [
            udf_result._repr_html_()
            if hasattr(udf_result, "_repr_html_")
            else repr(udf_result)
            for udf_result in self.udf_results
        ]
        return "<br><br>".join(result_reprs)

    model_config = ConfigDict(arbitrary_types_allowed=True)



from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Sequence, Union

from pydantic import BaseModel, ConfigDict, Field, StrictBool, StrictInt, StrictStr

from fused_batch._constants import DEFAULT_SHOW_TABLE_NAMES
from fused_batch.models.internal.dataset import DatasetInputV2Type
from fused_batch.models.udf import AnyBaseUdf

from .urls import DatasetUrl

# Note that instance types above 4xlarge are undocumented publicly
WHITELISTED_INSTANCE_TYPES = Literal[
    "m5.large",
    "m5.xlarge",
    "m5.2xlarge",
    "m5.4xlarge",
    "m5.8xlarge",
    "m5.12xlarge",
    "m5.16xlarge",
    "r5.large",
    "r5.xlarge",
    "r5.2xlarge",
    "r5.4xlarge",
    "r5.8xlarge",
    "r5.12xlarge",
    "r5.16xlarge",
]


class SampleMapRequest(BaseModel):
    file_id: Optional[str] = None
    chunk_id: Optional[int] = None
    n_rows: Optional[int] = None
    model_config = ConfigDict(populate_by_name=True)


class WholeFileSampleMapRequest(BaseModel):
    file_id: str = None
    n_rows: Optional[int] = None
    model_config = ConfigDict(populate_by_name=True)


class GetTableBboxRequest(BaseModel):
    path: str
    bbox_minx: float
    bbox_miny: float
    bbox_maxx: float
    bbox_maxy: float

    n_rows: Optional[int] = None
    clip: bool = True
    columns: Optional[List[str]] = None
    buffer: Optional[float] = None
    model_config = ConfigDict(populate_by_name=True)


class SampleJoinRequest(BaseModel):
    file_id: Optional[str] = None
    chunk_id: Optional[int] = None
    n_rows: Optional[int] = None
    model_config = ConfigDict(populate_by_name=True)


class SampleSingleFileJoinRequest(BaseModel):
    file_id: Optional[str] = None
    chunk_id: Optional[int] = None
    n_rows: Optional[int] = None
    model_config = ConfigDict(populate_by_name=True)


class StartJobRequest(BaseModel):
    region: Optional[StrictStr] = None
    instance_type: Optional[WHITELISTED_INSTANCE_TYPES] = None
    disk_size_gb: Optional[StrictInt] = None
    additional_env: Optional[List[StrictStr]] = Field(default_factory=list)
    backend: Optional[StrictStr] = None


class ListJobsRequest(BaseModel):
    skip: Optional[StrictInt] = None
    limit: Optional[StrictInt] = None


class UdfType(str, Enum):
    auto = "auto"
    vector_tile = "vector_tile"
    vector_single = "vector_single"
    raster = "raster"
    raster_single = "raster_single"
    app = "app"


class SaveUdfRequest(BaseModel):
    slug: Optional[str] = None
    udf_body: str
    udf_type: UdfType
    allow_public_read: Optional[bool] = None
    allow_public_list: Optional[bool] = None


class ListUdfsRequest(BaseModel):
    skip: Optional[StrictInt] = None
    limit: Optional[StrictInt] = None


class DebugDatasetRequest(BaseModel):
    path: DatasetUrl
    tables: Optional[Sequence[str]] = DEFAULT_SHOW_TABLE_NAMES
    dataset_config: Optional[Dict[str, Any]] = Field(default_factory=dict)
    app_config: Optional[Dict[str, Any]] = Field(default_factory=dict)


class DebugMultiDatasetRequestDataset(BaseModel):
    # TODO: Naming?
    path: DatasetUrl
    tables: Optional[Sequence[str]] = DEFAULT_SHOW_TABLE_NAMES
    dataset_config: Optional[Dict[str, Any]] = Field(default_factory=dict)


class DebugMultiDatasetRequest(BaseModel):
    datasets: Sequence[DebugMultiDatasetRequestDataset]
    app_config: Optional[Dict[str, Any]] = Field(default_factory=dict)


class MultiDebugSingleDatasetV2Request(BaseModel):
    urls: Sequence[DatasetUrl]
    operation: DatasetInputV2Type
    dataset_config: Optional[Dict[str, Any]] = Field(default_factory=dict)


class MultiDebugDatasetV2Request(BaseModel):
    app_config: Optional[Dict[str, Any]] = Field(default_factory=dict)
    datasets: List[MultiDebugSingleDatasetV2Request]


class GetPathRequest(BaseModel):
    path: DatasetUrl


class SignPathRequest(BaseModel):
    path: DatasetUrl


class ListPathRequest(BaseModel):
    path: DatasetUrl


class DeletePathRequest(BaseModel):
    path: DatasetUrl
    max_deletion_depth: Union[Optional[int], Literal["unlimited"]]


class ResolvePathRequest(BaseModel):
    path: DatasetUrl


class UploadRequest(BaseModel):
    path: DatasetUrl


class UploadTempRequest(BaseModel):
    extension: Optional[str] = None


class OpenDatasetRequest(BaseModel):
    path: DatasetUrl
    fetch_table_metadata: StrictBool
    fetch_samples: StrictBool = True


class OpenTableRequest(BaseModel):
    path: DatasetUrl


class OpenDatasetFolderRequest(BaseModel):
    path: DatasetUrl
    fetch_minimal_table_metadata: StrictBool
    fetch_table_metadata: StrictBool
    fetch_samples: StrictBool = True
    max_depth: Optional[StrictInt] = None


class ExplainUdfRequest(BaseModel):
    udf: AnyBaseUdf
    operation: Optional[str] = None


class GenerateUdfRequest(BaseModel):
    text: str
    operation: Optional[str] = None


class DebugUdfRequest(BaseModel):
    udf: AnyBaseUdf
    error: str


class ListUdfAccessTokensRequest(BaseModel):
    skip: Optional[StrictInt] = None
    limit: Optional[StrictInt] = None


class CreateUdfAccessTokenRequest(BaseModel):
    udf_email: Optional[str] = None
    udf_slug: Optional[str] = None
    udf_id: Optional[str] = None
    client_id: Optional[str] = None
    cache: bool = True
    metadata_json: Dict[str, Any] = {}
    enabled: bool = True


class UpdateUdfAccessTokenRequest(BaseModel):
    client_id: Optional[str] = None
    cache: Optional[bool] = None
    metadata_json: Optional[Dict[str, Any]] = None
    enabled: Optional[bool] = None



from __future__ import annotations

import base64
import functools
import hashlib
import inspect
import os
import pickle
import re
import shutil
import subprocess
import time
import types
from dataclasses import dataclass, fields
from datetime import datetime, timedelta, timezone
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Awaitable, Callable, Optional, TypeVar
from uuid import uuid4

from loguru import logger

from fused._environment import is_pyodide
from fused._options import options as OPTIONS

# Note: The entirety of this file, except with __future__ import, comes from job2
DEFAULT_CACHE_MAX_AGE = "12h"


@dataclass
class CacheLogEntry:
    expiration: datetime
    uuid: str
    status: str

    # TODO: Make class responsible for serialization to consolidate logic
    @classmethod
    def byte_length(cls):
        ex_length = 32  # datetime.isoformat
        uuid = 36  # str(uuid4())
        status = 1  # char
        commas = len(fields(cls)) - 1
        newline = 1
        return ex_length + uuid + status + commas + newline


# TODO: Possibly move all related methods to a class


def _serialize_code(code: types.CodeType) -> bytes:
    func_code_bytes = base64.b64encode(code.co_code)

    # Introduce variable values into hash.
    func_const_values = [
        # Embedded lambdas must be serialized, same as the overall function
        (
            _serialize_code(const)
            if isinstance(const, types.CodeType)
            else str(const).encode("utf-8")
        )
        for const in code.co_consts
    ]
    # Names must be serialized as otherwise referenced names (min vs max) would not
    # affect cache key.
    func_name_values = [name.encode("utf-8") for name in code.co_names]

    return b"".join(
        [
            func_code_bytes,
            *func_const_values,
            *func_name_values,
        ]
    )


def _serialize_function_defaults(func: Callable) -> bytes:
    parts = []

    if func.__defaults__ is not None:
        for d in func.__defaults__:
            parts.append(_hashify(d).encode("utf-8"))

    if func.__kwdefaults__ is not None:
        for key, value in func.__kwdefaults__.items():
            parts.append(f"{key}{_hashify(value)}".encode("utf-8"))

    return b"".join(parts)


def _hashify(func) -> str:
    hash_object = hashlib.sha256()
    try:
        if hasattr(func, "__fused_cached_fn"):
            return _hashify(func.__fused_cached_fn)
        elif callable(func):
            hash_object.update(_serialize_code(func.__code__))
            # Caution! The defaults and args do not go into the same part of the cache key!
            hash_object.update(_serialize_function_defaults(func))
        else:
            hash_object.update(str(func).encode("utf-8"))
        return hash_object.hexdigest()
    except Exception as e:
        logger.warning(f"Error Hashing {e}")
        return ""


def _parse_time_format(t: str | int) -> timedelta:
    if t == 0 or (isinstance(t, str) and t.strip() == "0"):
        return timedelta(0)

    if isinstance(t, int):
        t = f"{t}s"

    quantifier = re.match(r"(\d+)([smhd])$", t.strip().lower())
    if not quantifier:
        raise ValueError(
            f"Invalid time format {t!r}: Use a number followed by one of 's' (seconds), 'm' (minutes), 'h' (hours), or 'd' (days)."
        )
    value, unit = int(quantifier.group(1)), quantifier.group(2)
    if value < 0:
        raise ValueError(f"Time format {t!r} cannot be less than zero")

    delta_map = {
        "s": timedelta(seconds=value),
        "m": timedelta(minutes=value),
        "h": timedelta(hours=value),
        "d": timedelta(days=value),
    }

    return delta_map[unit]


def _format_now_with_delta(delta: timedelta) -> str:
    now = datetime.now(timezone.utc)
    return (now + delta).isoformat()


def _write_cache_file(
    func: Callable,
    args: tuple,
    kwargs: dict,
    cache_path: Path,
    uuid: str,
    expires_on: str,
    timeout_delta: timedelta,
):
    logger.debug(f"Caching {func.__qualname__} under {uuid}")
    in_prog_expires = _format_now_with_delta(timeout_delta)
    # Acquire lock and set status to in-progress
    _write_cache_log(cache_path, uuid, in_prog_expires, status="p")
    try:
        data = func(*args, **kwargs)
    except Exception as e:
        logger.debug(f"Error running {func.__qualname__}: {e}")
        # Release lock and set status to fail
        _write_cache_log(cache_path, uuid, expires_on, status="f")
        raise

    with open(cache_path / f"{uuid}.pickle", "wb") as f:
        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    # Release lock and set status to done
    _write_cache_log(cache_path, uuid, expires_on, status="d")
    return data


async def _write_cache_file_async(
    func: Callable,
    args: tuple,
    kwargs: dict,
    cache_path: Path,
    uuid: str,
    expires_on: str,
    timeout_delta: timedelta,
):
    logger.debug(f"Caching {func.__qualname__} under {uuid}")
    in_prog_expires = _format_now_with_delta(timeout_delta)
    # Acquire lock and set status to in-progress
    _write_cache_log(cache_path, uuid, in_prog_expires, status="p")
    try:
        data = await func(*args, **kwargs)
    except Exception as e:
        logger.debug(f"Error running {func.__qualname__}: {e}")
        # Release lock and set status to fail
        _write_cache_log(cache_path, uuid, expires_on, status="f")
        raise

    with open(cache_path / f"{uuid}.pickle", "wb") as f:
        pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)

    # Release lock and set status to done
    _write_cache_log(cache_path, uuid, expires_on, status="d")
    return data


def _write_cache_log(cache_path: Path, uuid: str, expires_on: str, status: str):
    log_file = cache_path / "index.log"
    temp_log = cache_path / f"{uuid}.log"
    shutil.copy(log_file, temp_log)
    with open(temp_log, "a") as f:
        f.write(f"{expires_on},{uuid},{status}\n")
    os.rename(temp_log, log_file)
    logger.debug(f"Wrote {uuid} to cache log {log_file} with status {status}")


def _read_cache_log(cache_path: Path) -> CacheLogEntry | None:
    log_file = cache_path / "index.log"
    # Ensure log file exists
    log_file.touch()
    if is_pyodide():
        try:
            with open(log_file, "rb") as logfile:
                logfile.seek(CacheLogEntry.byte_length() * -1, 2)
                recent_cache_entry = logfile.read().strip().decode("utf8")
        except OSError as e:
            # file is empty
            if "invalid argument" in str(e).lower():
                recent_cache_entry = None
            else:
                logger.debug(f"Reading log file failed: {e}")
                raise

    else:
        try:
            # TODO: Replace tail with builtin file reads and update log file to use fixed byte length record sizes
            # Strip to ensure we don't break if an empty line exists
            recent_cache_entry = subprocess.check_output(
                ["tail", "-1", str(log_file)], encoding="utf8"
            ).strip()
        except CalledProcessError as e:
            # Something wrong with the file or executing command
            logger.debug(f"Reading log file failed: {e}")
            raise

    if recent_cache_entry:
        try:
            expiration, uuid, status = (
                d.strip() for d in recent_cache_entry.split(",")
            )
            logger.debug(
                f"Recent cache entry found in {log_file}: {expiration!r}, {uuid!r}, {status!r}"
            )
            return CacheLogEntry(datetime.fromisoformat(expiration), uuid, status)
        except (TypeError, ValueError) as e:
            logger.debug(
                f"Issues with parsing last line in log {e}. Truncating logfile {log_file}"
            )
            # Any issue with parsing the last line. This could mean the file is corrupted, therefore we clear the file.
            with open(log_file, "w") as logfile:
                logfile.truncate()

    return None


def _read_cache_file(cache_path: Path, uuid: str):
    with open(cache_path / f"{uuid}.pickle", "rb") as f:
        data = pickle.load(f)
        return data


def _cache(
    func: Callable,
    *args,
    cache_max_age: str | int = DEFAULT_CACHE_MAX_AGE,
    path: str = "tmp",
    concurrent_lock_timeout: str | int = 120,
    reset: bool = False,
    **kwargs,
) -> Any:
    """Internal method used by cache decorator to cache a function's pickleable response to a file.

    Args:
        func: The decorated function.
        *args: Positional arguments to the decorated function.
        cache_max_age: A string with a numbered component and units. Supported units are seconds (s), minutes (m),
            hours (h), and days (d) (e.g. "48h", "10s", etc.).
        path: Folder to append to the configured cache directory.
        concurrent_lock_timeout: Max amount of time for concurrent calls to wait for the decorated function
            to finish execution and to write the cache file currently being written by another concurrent call.
            Waiting will end before the timeout if a finished cache file is detected after reading the most recent entry
            in the log. Otherwise, after the timeout, it will either find a finished cached file to read, or it will
            write a new one if a cache file has not been recorded in the log.
        **kwargs: Keyword arguments to the decorated function.

    Returns:
        Any: Result from the cached function
    """
    # Calculate expires_on
    max_age_delta = _parse_time_format(cache_max_age)
    timeout_delta = _parse_time_format(concurrent_lock_timeout)

    try:
        path = path.strip("/")
        # TODO: consider udf name in path once available from Fused global context
        path = OPTIONS.cache_directory / path
        # TODO: ignore `_`

        # 1. Hashify function
        id = _hashify(func)

        # 2. Hashify args
        for v in args:
            id += "_" + _hashify(v)

        # 3. Hashify kwargs
        for k in kwargs:
            id += k + _hashify(kwargs[k])

        # 4. Hashify cache_max_age
        id += _hashify(max_age_delta)

        # 5. Hashify composite id
        id = _hashify(id)

        path_dir = path / f"data_{id}"

        path_dir.mkdir(parents=True, exist_ok=True)

        cache_entry = _read_cache_log(path_dir)
        new_cache_uuid = str(uuid4())
        expires_on = _format_now_with_delta(max_age_delta)
        timeout_seconds = timeout_delta.total_seconds()
        if not reset and cache_entry:
            now = datetime.now(timezone.utc)
            wait_time_seconds = 1
            max_retries = timeout_seconds // wait_time_seconds
            retries = 1
            while (
                cache_entry.status == "p"
                and now < cache_entry.expiration
                and retries < max_retries
            ):
                logger.debug(
                    f"Sleeping. Cache entry reporting {cache_entry.status}, {now} < {cache_entry.expiration}, "
                    f"current write might be in progress. Retry [{retries}/{max_retries}]"
                )
                time.sleep(wait_time_seconds)
                cache_entry = _read_cache_log(path_dir)
                retries += 1

            logger.debug(
                f"Detected status is {cache_entry.status} and expiration is {cache_entry.expiration}"
            )
            if cache_entry.status == "d" and now < cache_entry.expiration:
                logger.debug(
                    f"Reading cache for {func.__qualname__} under {cache_entry.uuid}"
                )
                # Cache exists and valid
                return _read_cache_file(path_dir, cache_entry.uuid)
            else:
                return _write_cache_file(
                    func,
                    args,
                    kwargs,
                    path_dir,
                    new_cache_uuid,
                    expires_on,
                    timeout_delta,
                )

        else:
            return _write_cache_file(
                func, args, kwargs, path_dir, new_cache_uuid, expires_on, timeout_delta
            )

    except Exception as e:
        logger.debug(f"Error Caching {e}")
        raise e


async def _cache_async(
    func: Callable,
    *args,
    cache_max_age: str | int = DEFAULT_CACHE_MAX_AGE,
    path: str = "tmp",
    concurrent_lock_timeout: str | int = 120,
    reset: bool = False,
    **kwargs,
) -> Any:
    """Async internal method used by cache decorator to cache a function's pickleable response to a file.

    Args:
        func: The decorated function.
        *args: Positional arguments to the decorated function.
        cache_max_age: A string with a numbered component and units. Supported units are seconds (s), minutes (m),
            hours (h), and days (d) (e.g. "48h", "10s", etc.).
        path: Folder to append to the configured cache directory.
        concurrent_lock_timeout: Max amount of time for concurrent calls to wait for the decorated function
            to finish execution and to write the cache file currently being written by another concurrent call.
            Waiting will end before the timeout if a finished cache file is detected after reading the most recent entry
            in the log. Otherwise, after the timeout, it will either find a finished cached file to read, or it will
            write a new one if a cache file has not been recorded in the log.
        **kwargs: Keyword arguments to the decorated function.

    Returns:
        Any: Result from the cached function
    """
    # Calculate expires_on
    max_age_delta = _parse_time_format(cache_max_age)
    timeout_delta = _parse_time_format(concurrent_lock_timeout)

    try:
        path = path.strip("/")
        # TODO: consider udf name in path once available from Fused global context
        path = OPTIONS.cache_directory / path
        # TODO: ignore `_`

        # 1. Hashify function
        id = _hashify(func)

        # 2. Hashify args
        for v in args:
            id += "_" + _hashify(v)

        # 3. Hashify kwargs
        for k in kwargs:
            id += k + _hashify(kwargs[k])

        # 4. Hashify cache_max_age
        id += _hashify(max_age_delta)

        # 5. Hashify composite id
        id = _hashify(id)

        path_dir = path / f"data_{id}"

        path_dir.mkdir(parents=True, exist_ok=True)

        cache_entry = _read_cache_log(path_dir)
        new_cache_uuid = str(uuid4())
        expires_on = _format_now_with_delta(max_age_delta)
        timeout_seconds = timeout_delta.total_seconds()
        if not reset and cache_entry:
            now = datetime.now(timezone.utc)
            wait_time_seconds = 1
            max_retries = timeout_seconds // wait_time_seconds
            retries = 1
            while (
                cache_entry.status == "p"
                and now < cache_entry.expiration
                and retries < max_retries
            ):
                logger.debug(
                    f"Sleeping. Cache entry reporting {cache_entry.status}, {now} < {cache_entry.expiration}, "
                    f"current write might be in progress. Retry [{retries}/{max_retries}]"
                )
                time.sleep(wait_time_seconds)
                cache_entry = _read_cache_log(path_dir)
                retries += 1

            logger.debug(
                f"Detected status is {cache_entry.status} and expiration is {cache_entry.expiration}"
            )
            if cache_entry.status == "d" and now < cache_entry.expiration:
                logger.debug(
                    f"Reading cache for {func.__qualname__} under {cache_entry.uuid}"
                )
                # Cache exists and valid
                return _read_cache_file(path_dir, cache_entry.uuid)
            else:
                return await _write_cache_file_async(
                    func,
                    args,
                    kwargs,
                    path_dir,
                    new_cache_uuid,
                    expires_on,
                    timeout_delta,
                )

        else:
            return await _write_cache_file_async(
                func, args, kwargs, path_dir, new_cache_uuid, expires_on, timeout_delta
            )

    except Exception as e:
        logger.debug(f"Error Caching {e}")
        raise e


def _cache_internal(func, **decorator_kwargs):
    def decorator(func):
        if inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def wrapper_async(*args, **kwargs):
                # Allow passing cache kwargs through fn kwargs. Depends on all cache kwargs to be specified from cache
                # decorator
                for k in decorator_kwargs:
                    if k in kwargs:
                        decorator_kwargs[k] = kwargs.pop(k)

                return await _cache_async(
                    func,
                    *args,
                    **decorator_kwargs,
                    **kwargs,
                )

            wrapper_async.__fused_cached_fn = func

            return wrapper_async
        else:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # Allow passing cache kwargs through fn kwargs. Depends on all cache kwargs to be specified from cache
                # decorator
                for k in decorator_kwargs:
                    if k in kwargs:
                        decorator_kwargs[k] = kwargs.pop(k)

                return _cache(
                    func,
                    *args,
                    **decorator_kwargs,
                    **kwargs,
                )

            wrapper.__fused_cached_fn = func

            return wrapper

    if callable(func):  # w/o args
        return decorator(func)
    else:  # w/ args
        return decorator


def cache(
    func: Optional[Callable[..., Any]] = None,
    cache_max_age: str | int = DEFAULT_CACHE_MAX_AGE,
    path: str = "tmp",
    concurrent_lock_timeout: str | int = 120,
    reset: bool = False,
) -> Callable[..., Any]:
    """Decorator to cache the return value of a function.

    This function serves as a decorator that can be applied to any function
    to cache its return values. The cache behavior can be customized through
    keyword arguments.

    Args:
        func (Callable, optional): The function to be decorated. If None, this
            returns a partial decorator with the passed keyword arguments.
        cache_max_age: A string with a numbered component and units. Supported units are seconds (s), minutes (m), hours (h), and
            days (d) (e.g. "48h", "10s", etc.).
        path: Folder to append to the configured cache directory.
        concurrent_lock_timeout: Max amount of time in seconds for subsequent concurrent calls to wait for a previous
            concurrent call to finish execution and to write the cache file.

    Returns:
        Callable: A decorator that, when applied to a function, caches its
        return values according to the specified keyword arguments.

    Examples:

        Use the `@cache` decorator to cache the return value of a function in a custom path.

        ```py
        @cache(path="/tmp/custom_path/")
        def expensive_function():
            # Function implementation goes here
            return result
        ```

        If the output of a cached function changes, for example if remote data is modified,
        it can be reset by running the function with the `reset` keyword argument. Afterward,
        the argument can be cleared.

        ```py
        @cache(path="/tmp/custom_path/", reset=True)
        def expensive_function():
            # Function implementation goes here
            return result
        ```
    """
    return _cache_internal(
        func=func,
        cache_max_age=cache_max_age,
        path=path,
        concurrent_lock_timeout=concurrent_lock_timeout,
        reset=reset,
    )


T = TypeVar("T")


def cache_call(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """Directly calls a function with caching.

    This function directly calls the provided function with the given arguments
    and keyword arguments, caching its return value. The cache used depends on
    the implementation of the `_cache` function.

    Args:
        func (Callable): The function to call and cache its result.
        *args: Positional arguments to pass to the function.
        **kwargs: Keyword arguments to pass to the function.

    Returns:
        The cached return value of the function.

    Raises:
        Exception: Propagates any exception raised by the function being called
        or the caching mechanism.
    """
    return _cache(func, *args, **kwargs)


async def cache_call_async(
    func: Callable[..., Awaitable[T]], *args: Any, **kwargs: Any
) -> T:
    """Asynchronously calls a function with caching.

    Similar to `cache_call`, but for asynchronous functions. This function
    awaits the provided async function, caches its return value, and then
    returns it. The specifics of the caching mechanism depend on the
    implementation of `_cache_async`.

    Args:
        func (Callable): The asynchronous function to call and cache its result.
        *args: Positional arguments to pass to the async function.
        **kwargs: Keyword arguments to pass to the async function.

    Returns:
        The cached return value of the async function.

    Raises:
        Exception: Propagates any exception raised by the async function being
        called or the caching mechanism.

    Examples:
        async def fetch_data(param):
            # Async function implementation goes here
            return data

        # Usage
        result = await cache_call_async(fetch_data, 'example_param')
    """
    return await _cache_async(func, *args, **kwargs)
